-- dia2postgresql.pl inh library
--
-- Handles inheritance and primary key enforcement on a table hierarchy
--
-- Copyright (C) 2002 E.BENOIT   <tseeker@nocternity.net>
--
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--


--
-- Create the inheritance cache table (d2p_inh_cache)
--
CREATE SEQUENCE d2p_inh_cache_id_seq;

CREATE TABLE d2p_inh_cache(
	id		INT NOT NULL,
	ancestor	TEXT NOT NULL,
	child		TEXT NOT NULL,
	real		BOOL NOT NULL,
	base		BOOL NOT NULL DEFAULT false,
	PRIMARY KEY(id),
	UNIQUE(ancestor,child)
);


--
-- Create the inheritance ignore list ; inheritance relations listed in this
-- table will not be handled by the triggers
--
CREATE TABLE d2p_inh_ignore(
	ancestor	TEXT NOT NULL,
	child		TEXT NOT NULL,
	PRIMARY KEY(ancestor,child)
);


--
-- Function that creates a class' inheritance tree and global children list
--
CREATE FUNCTION f_d2p_inh_addcache (INT,TEXT,TEXT,BOOL,BOOL) RETURNS INT AS '
DECLARE
	tmp	TEXT;
	t2	TEXT;
	c	refcursor;
	mi	INT;
	i2	INT;
BEGIN
	INSERT INTO d2p_inh_cache(id, ancestor, child, real, base) VALUES ($1,$2,$3,$4,$5);
	IF $4 THEN
		SELECT INTO tmp ancestor FROM d2p_inh_cache WHERE ancestor = $3 GROUP BY ancestor;
		IF tmp IS NULL THEN
			OPEN c FOR SELECT c2.relname
				FROM pg_class c2, pg_inherits i, pg_class c1, pg_tables t
				WHERE c2.oid = i.inhrelid
				 AND i.inhparent = c1.oid
				 AND c1.relname = t.tablename
				 AND t.tablename = $3;

			LOOP
				FETCH c INTO tmp;
				IF tmp IS NULL THEN
					EXIT;
				END IF;

				SELECT INTO t2 ancestor FROM d2p_inh_ignore WHERE ancestor = $3 AND child = tmp;
				IF t2 IS NULL THEN
					SELECT INTO mi nextval(''d2p_inh_cache_id_seq'');
					SELECT INTO i2 f_d2p_inh_addcache (mi, $3, tmp, true, false);
				END IF;
			END LOOP;

			CLOSE c;
		END IF;

		OPEN c FOR SELECT child FROM d2p_inh_cache WHERE ancestor = $3;
		LOOP
			FETCH c INTO tmp;
			IF tmp IS NULL THEN
				EXIT;
			END IF;
			SELECT INTO t2 ancestor FROM d2p_inh_ignore WHERE ancestor = $2 AND child = tmp;
			IF t2 IS NULL THEN
				SELECT INTO mi nextval(''d2p_inh_cache_id_seq'');
				SELECT INTO tmp f_d2p_inh_addcache (mi, $2, tmp, false, $5);
			END IF;
		END LOOP;
		CLOSE c;
	END IF;
	RETURN	0;
END' LANGUAGE 'plpgsql';



--
-- Function that creates the inheritance cache (d2p_inh_cache)
--
CREATE FUNCTION f_d2p_inh_createcache () RETURNS BOOL AS '
DECLARE
	rt	RECORD;
	cx	refcursor;
	c3	refcursor;
	ch	TEXT;
	i	INT;
	i3	INT;
	tmp	TEXT;
BEGIN
	OPEN c3 FOR SELECT t.tablename AS n, c.oid AS o
		FROM	pg_class c, pg_tables t
		WHERE	(t.tablename = c.relname
		 AND	t.tablename NOT LIKE ''pg_%''
		 AND	t.tablename NOT LIKE ''d2p_%''
		 AND	c.relhaspkey
		 AND	c.oid NOT IN	(SELECT inhrelid
						FROM pg_inherits
						GROUP BY inhrelid
					)
		);

	LOOP
		FETCH c3 INTO rt;
		IF NOT FOUND THEN
			EXIT;
		END IF;

		OPEN cx FOR SELECT c.relname
				FROM	pg_class c, pg_inherits i2
				WHERE	(c.oid = i2.inhrelid AND i2.inhparent = rt.o);
		LOOP
			FETCH cx INTO ch;
			IF ch IS NULL THEN
				EXIT;
			END IF;
			SELECT INTO tmp ancestor FROM d2p_inh_ignore WHERE ancestor = rt.n AND child = ch;
			IF tmp IS NULL THEN
				SELECT INTO i nextval(''d2p_inh_cache_id_seq'');
				SELECT INTO i3 f_d2p_inh_addcache (i, rt.n, ch, true, true);
			END IF;
		END LOOP;
		CLOSE cx;
	END LOOP;

	CLOSE c3;

	RETURN TRUE;
END' LANGUAGE 'plpgsql';


--
-- Function that creates the functions and triggers
--
CREATE FUNCTION f_d2p_inh_createft () RETURNS BOOL AS '
DECLARE
	rt	RECORD;
	i2	INT;
	na	TEXT;
	tmp	TEXT;
	toid	OID;
	tn	TEXT;
	cx	refcursor;
	c3	refcursor;
BEGIN
	OPEN c3 FOR SELECT t.tablename AS n, c.oid AS o
		FROM	pg_class c, pg_tables t
		WHERE	(t.tablename = c.relname
		 AND	t.tablename NOT LIKE ''pg_%''
		 AND	t.tablename NOT LIKE ''d2p_%''
		 AND	c.relhaspkey
		 AND	c.oid NOT IN	(SELECT inhrelid
						FROM pg_inherits
						GROUP BY inhrelid
					)
		);

	LOOP
		-- Create functions code
		FETCH c3 INTO rt;
		IF NOT FOUND THEN
			EXIT;
		END IF;

		SELECT INTO na, toid attname, atttypid FROM pg_attribute WHERE attrelid = rt.o AND attnum IN (
			SELECT indkey[0] FROM pg_index WHERE indrelid = rt.o AND indisprimary
		);
		IF na IS NULL THEN
			CLOSE c3;
			RAISE EXCEPTION ''Primary key not found for table %'', rt.n;
		END IF;

		SELECT INTO tn typname FROM pg_type WHERE oid = toid;
		IF tn IS NULL THEN
			CLOSE c3;
			RAISE EXCEPTION ''Type not found for % primary key'', rt.n;
		END IF;

		SELECT INTO i2 min(id) FROM d2p_inh_cache WHERE ancestor = rt.n;
		IF i2 IS NOT NULL THEN
			-- Create trigger functions
			tmp := ''CREATE FUNCTION f_d2p_inh_auto_in_'' || i2 || '' () RETURNS OPAQUE AS ''''
				DECLARE
					t	'' || tn || '';
				BEGIN
					SELECT INTO t "'' || na || ''" FROM "'' || rt.n || ''" WHERE "'' || na || ''" = NEW."'' || na || ''";
					IF t IS NOT NULL THEN
						RAISE EXCEPTION ''''''''Primary key value % already exists in '' || rt.n || '''''''''', t;
					END IF;
					RETURN NEW;
				END'''' LANGUAGE ''''plpgsql'''';'';
			EXECUTE tmp;
			tmp := ''CREATE FUNCTION f_d2p_inh_auto_up_'' || i2 || '' () RETURNS OPAQUE AS ''''
				DECLARE
					t	'' || tn || '';
				BEGIN
					IF OLD.'' || na || '' IS NOT NULL AND NEW.'' || na || '' = OLD.'' ||na|| '' THEN
						RETURN NEW;
					END IF;
					SELECT INTO t "'' || na || ''" FROM "'' || rt.n || ''" WHERE "'' || na || ''" = NEW."'' || na || ''";
					IF t IS NOT NULL THEN
						RAISE EXCEPTION ''''''''Primary key value % already exists in '' || rt.n || '''''''''', t;
					END IF;
					RETURN NEW;
				END'''' LANGUAGE ''''plpgsql'''';'';
			EXECUTE tmp;

			-- Create helper function
			tmp := ''CREATE FUNCTION "'' || rt.n || ''_keytotable" ('' || tn || '') RETURNS TEXT AS ''''
				DECLARE
					t	TEXT;
				BEGIN
					SELECT INTO t c.relname FROM pg_class c, "'' || rt.n || ''" d
						WHERE c.oid = d.tableoid AND d."'' || na || ''" = $1;
					RETURN t;
				END'''' LANGUAGE ''''plpgsql'''';'';
			EXECUTE tmp;

			-- Create triggers
			OPEN cx FOR SELECT child FROM d2p_inh_cache WHERE ancestor = rt.n;
			LOOP
				FETCH cx INTO tn;
				IF tn IS NULL THEN
					EXIT;
				END IF;
				tmp := ''CREATE TRIGGER "t_d2p_inh_auto_in_'' || i2 || ''" BEFORE INSERT ON "'' || tn || ''"
					FOR EACH ROW EXECUTE PROCEDURE "f_d2p_inh_auto_in_'' || i2 || ''" ();'';
				EXECUTE tmp;
				tmp := ''CREATE TRIGGER "t_d2p_inh_auto_up_'' || i2 || ''" BEFORE UPDATE ON "'' || tn || ''"
					FOR EACH ROW EXECUTE PROCEDURE "f_d2p_inh_auto_up_'' || i2 || ''" ();'';
				EXECUTE tmp;
			END LOOP;
			CLOSE cx;
			tmp := ''CREATE TRIGGER "t_d2p_inh_auto_in_'' || i2 || ''" BEFORE INSERT ON "'' || rt.n || ''"
				FOR EACH ROW EXECUTE PROCEDURE  "f_d2p_inh_auto_in_'' || i2 || ''" ();'';
			EXECUTE tmp;
			tmp := ''CREATE TRIGGER "t_d2p_inh_auto_up_'' || i2 || ''" BEFORE UPDATE ON "'' || rt.n || ''"
				FOR EACH ROW EXECUTE PROCEDURE  "f_d2p_inh_auto_up_'' || i2 || ''" ();'';
			EXECUTE tmp;
		END IF;
	END LOOP;

	CLOSE c3;

	RETURN TRUE;
END' LANGUAGE 'plpgsql';


--
-- Function that removes tables that no longer exist from the cache
--
CREATE FUNCTION f_d2p_inh_deldropped () RETURNS BOOL AS '
DECLARE
	cur	refcursor;
	tn	TEXT;
	t2	TEXT;
	toid	OID;
	ba	BOOL;
	i2	INT;
BEGIN
	OPEN cur FOR SELECT ancestor FROM d2p_inh_cache UNION SELECT CHILD FROM d2p_inh_cache;
	LOOP
		FETCH cur INTO tn;
		IF tn IS NULL THEN
			EXIT;
		END IF;

		SELECT INTO toid c.oid FROM pg_class c, pg_tables t WHERE c.relname = t.tablename AND t.tablename = tn;
		IF toid IS NULL THEN
			SELECT INTO i2 min(id) FROM d2p_inh_cache WHERE ancestor = tn;
			IF i2 IS NOT NULL THEN
				SELECT INTO ba base FROM d2p_inh_cache WHERE ancestor = tn AND base AND id = i2;
				IF ba THEN
					EXECUTE ''DROP FUNCTION "f_d2p_inh_auto_in_'' || i2 || ''" () ;'';
					EXECUTE ''DROP FUNCTION "f_d2p_inh_auto_up_'' || i2 || ''" () ;'';
					EXECUTE ''DROP FUNCTION "'' || tn || ''_keytotable" () ;'';
					SELECT INTO t2 t.typname FROM pg_proc p, pg_type t WHERE p.proname = tn || ''_keytotable'' AND t.oid = p.proargtypes[0];
					EXECUTE ''DROP FUNCTION "'' || tn || ''_keytotable" ('' || t2 || '') ;'';
				END IF;
			END IF;
			DELETE FROM d2p_inh_cache WHERE ancestor = tn OR child = tn;
		END IF;
	END LOOP;
	CLOSE cur;
	RETURN true;
END' LANGUAGE 'plpgsql';


--
-- Function that deletes the existing inh-generated triggers and functions
--
CREATE FUNCTION f_d2p_inh_deltf() RETURNS BOOL AS '
DECLARE
        cur     refcursor;
	rt	RECORD;
	oa	TEXT;
	ca	TEXT;
	tn	TEXT;
	fn	TEXT;
	i2	INT;
BEGIN
	OPEN cur FOR SELECT ancestor, child FROM d2p_inh_cache WHERE base ORDER BY ancestor;
	ca := '''';
	LOOP
		FETCH cur INTO rt;
		IF NOT FOUND THEN
			EXIT;
		END IF;

		oa := ca;
		ca := rt.ancestor;
		IF oa <> '''' AND ca <> oa THEN
			EXECUTE ''DROP TRIGGER "t_d2p_inh_auto_in_'' || i2 || ''" ON "'' || oa || ''";'';
			EXECUTE ''DROP TRIGGER "t_d2p_inh_auto_up_'' || i2 || ''" ON "'' || oa || ''";'';
			EXECUTE ''DROP FUNCTION "f_d2p_inh_auto_in_'' || i2 || ''" () ;'';
			EXECUTE ''DROP FUNCTION "f_d2p_inh_auto_up_'' || i2 || ''" () ;'';

			fn := substring((oa || ''_keytotable'') from 1 for 31);
			SELECT INTO tn t.typname FROM pg_proc p, pg_type t WHERE p.proname = fn AND t.oid = p.proargtypes[0];
			EXECUTE ''DROP FUNCTION "'' || fn || ''" ('' || tn || '') ;'';
			SELECT INTO i2 min(id) FROM d2p_inh_cache WHERE ancestor = ca;
		ELSIF oa = '''' THEN
			SELECT INTO i2 min(id) FROM d2p_inh_cache WHERE ancestor = ca;
		END IF;

		EXECUTE ''DROP TRIGGER "t_d2p_inh_auto_in_'' || i2 || ''" ON "'' || rt.child || ''";'';
		EXECUTE ''DROP TRIGGER "t_d2p_inh_auto_up_'' || i2 || ''" ON "'' || rt.child || ''";'';
	END LOOP;

	IF ca <> '''' THEN
		EXECUTE ''DROP TRIGGER "t_d2p_inh_auto_in_'' || i2 || ''" ON "'' || ca || ''";'';
		EXECUTE ''DROP TRIGGER "t_d2p_inh_auto_up_'' || i2 || ''" ON "'' || ca || ''";'';
		EXECUTE ''DROP FUNCTION "f_d2p_inh_auto_in_'' || i2 || ''" () ;'';
		EXECUTE ''DROP FUNCTION "f_d2p_inh_auto_up_'' || i2 || ''" () ;'';
		SELECT INTO tn t.typname FROM pg_proc p, pg_type t WHERE p.proname = ca || ''_keytotable'' AND t.oid = p.proargtypes[0];
		EXECUTE ''DROP FUNCTION "'' || ca || ''_keytotable" ('' || tn || '') ;'';
	END IF;

	CLOSE cur;
	RETURN true;
END' LANGUAGE 'plpgsql';


--
-- Function that must be called to update the cache, triggers and functions
--
CREATE FUNCTION f_d2p_inh_update () RETURNS BOOL AS '
DECLARE
	junk	BOOL;
BEGIN
	-- Delete dropped tables
	SELECT INTO junk f_d2p_inh_deldropped ();
	-- Drop triggers and functions
	SELECT INTO junk f_d2p_inh_deltf ();
	-- Delete old cache
	DELETE FROM d2p_inh_cache;
	-- Create new cache
	SELECT INTO junk f_d2p_inh_createcache ();
	-- Create triggers
	SELECT INTO junk f_d2p_inh_createft ();
	RETURN TRUE;
END' LANGUAGE 'plpgsql';

