-- dia2postgresql.pl ref library
--
-- Handles references between tables, with support for inheritance
--
-- Copyright (C) 2002 E.BENOIT   <tseeker@nocternity.net>
--
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--


--
-- Create d2p_ref tables
--
CREATE TABLE d2p_ref_list(
	id		SERIAL PRIMARY KEY,
	src_table	TEXT NOT NULL,
	src_field	TEXT NOT NULL,
	dest_table	TEXT NOT NULL,
	dest_field	TEXT NOT NULL,
	del_cascade	BOOL NOT NULL DEFAULT false,
	upd_cascade	BOOL NOT NULL DEFAULT false,
	UNIQUE(src_table, src_field)
);
CREATE TABLE d2p_ref_list_upd(
	id		SERIAL PRIMARY KEY,
	src_table	TEXT NOT NULL,
	src_field	TEXT NOT NULL,
	dest_table	TEXT NOT NULL,
	dest_field	TEXT NOT NULL,
	UNIQUE(src_table, src_field)
);

CREATE TABLE d2p_ref_functions(
	idref		INT REFERENCES d2p_ref_list(id) NOT NULL,
	otype		CHAR(2) NOT NULL,
	PRIMARY KEY(idref, otype)
);

CREATE TABLE d2p_ref_triggers(
	idref		INT REFERENCES d2p_ref_list(id) NOT NULL,
	otype		CHAR(2) NOT NULL,
	relname		TEXT NOT NULL,
	PRIMARY KEY(idref, otype, relname)
);


--
-- Deny update on d2p_ref functions and triggers tables
--
CREATE FUNCTION f_d2p_ref_denied () RETURNS OPAQUE AS '
BEGIN
	RAISE EXCEPTION ''Operation denied by dia2postgresql trigger'';
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_upddenied BEFORE UPDATE ON d2p_ref_functions
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_denied ();
CREATE TRIGGER t_d2p_ref_upddenied BEFORE UPDATE ON d2p_ref_triggers
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_denied ();

--
-- Trigger that checks for cascade mode changes in d2p_ref_list
--
CREATE FUNCTION f_d2p_ref_check_updlist () RETURNS OPAQUE AS '
BEGIN
	IF NEW.id <> OLD.id OR NEW.src_table <> OLD.src_table OR NEW.src_field <> OLD.src_field
		OR NEW.dest_table <> OLD.dest_table OR NEW.dest_field <> OLD.dest_field
	THEN
		RAISE EXCEPTION ''You are not allowed to modify fields other than del_ and upd_cascade'';
	END IF;

	RETURN NEW;
END' LANGUAGE 'plpgsql';
CREATE TRIGGER t_d2p_ref_check_update BEFORE UPDATE ON d2p_ref_list
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_check_updlist ();


--
-- Trigger that deletes then reinserts rows from/into d2p_ref_functions when cascade mode changes happen
--
CREATE FUNCTION f_d2p_ref_updlist () RETURNS OPAQUE AS '
BEGIN
	DELETE FROM d2p_ref_functions WHERE idref = NEW.id;
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''in'', NEW.id);
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''us'', NEW.id);
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''ud'', NEW.id);
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''rm'', NEW.id);

	RETURN NEW;
END' LANGUAGE 'plpgsql';
CREATE TRIGGER t_d2p_ref_do_update AFTER UPDATE ON d2p_ref_list
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_updlist ();


--
-- Trigger that checks the operation type for functions and triggers
--
CREATE FUNCTION f_d2p_ref_check_otype () RETURNS OPAQUE AS '
BEGIN
	IF NOT NEW.otype IN (''in'',''ud'',''us'',''rm'') THEN
		RAISE EXCEPTION ''Invalid operation type : %'', NEW.otype;
	END IF;
	RETURN NEW;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_check_otype BEFORE INSERT ON d2p_ref_functions
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_check_otype ();
CREATE TRIGGER t_d2p_ref_check_otype BEFORE INSERT ON d2p_ref_triggers
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_check_otype ();


--
-- Trigger that insert the appropriate rows into d2p_ref_functions on insertion into d2p_ref_list
--
CREATE FUNCTION f_d2p_ref_addref () RETURNS OPAQUE AS '
BEGIN
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''in'', NEW.id);
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''us'', NEW.id);
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''ud'', NEW.id);
	INSERT INTO d2p_ref_functions(otype, idref) VALUES(''rm'', NEW.id);
	RETURN	NULL;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_addref AFTER INSERT ON d2p_ref_list
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_addref ();


--
-- Trigger that deletes lines from d2p_ref_functions on deletion from d2p_ref_list
--
CREATE FUNCTION f_d2p_ref_delref () RETURNS OPAQUE AS '
BEGIN
	DELETE FROM d2p_ref_functions WHERE idref = OLD.id;
	RETURN	OLD;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_delref BEFORE DELETE ON d2p_ref_list
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_delref ();


--
-- Function that generates the code for the insertion check trigger function for a reference
--
CREATE FUNCTION f_d2p_ref_code_f_in(INT) RETURNS TEXT AS '
DECLARE
	idref	ALIAS FOR $1;
	st	TEXT;
	sf	TEXT;
	dt	TEXT;
	df	TEXT;
	tn	TEXT;
	tmp	TEXT;
BEGIN
	SELECT INTO st, sf, dt, df 
		src_table, src_field, dest_table, dest_field FROM d2p_ref_list WHERE id = idref;

	IF df IS NULL THEN
		RAISE EXCEPTION ''Reference % not found in d2p_ref_list'', $1;
	END IF;

	SELECT INTO tn typname FROM pg_type t, pg_attribute a, pg_class c
		WHERE t.oid = a.atttypid AND a.attrelid = c.oid AND a.attname = sf AND c.relname = st;
	IF tn IS NULL THEN
		tn := st || ''.'' || sf;
		RAISE EXCEPTION ''Type not found for %'', tn;
	END IF;

	tmp := ''DECLARE
		c CURSOR FOR SELECT '' || df || '' FROM '' || dt || '';
		id '' ||tn||'';
	BEGIN
		IF NEW.'' || sf || '' IS NULL THEN
			RETURN NEW;
		END IF;

		OPEN c;
		LOOP
			FETCH c INTO id;
			IF id IS NULL THEN
				EXIT;
			END IF;
			IF NEW.'' || sf || '' = id THEN
				CLOSE c;
				RETURN NEW;
			END IF;
		END LOOP;
		CLOSE c;
		RAISE EXCEPTION ''''''''Value % not found for '' || dt || ''.'' || df || '''''''''', NEW.'' || sf || '';
	END'';
	RETURN tmp;
END' LANGUAGE 'plpgsql';


--
-- Function that generates the code for the destination table delete check trigger function for a reference
--
CREATE FUNCTION f_d2p_ref_code_f_ud(INT) RETURNS TEXT AS '
DECLARE
	idref	ALIAS FOR $1;
	st	TEXT;
	sf	TEXT;
	dt	TEXT;
	tn	TEXT;
	df	TEXT;
	tmp	TEXT;
	uc	BOOL;
BEGIN
	SELECT INTO st, sf, dt, df, uc
		src_table, src_field, dest_table, dest_field, upd_cascade FROM d2p_ref_list WHERE id = idref;

	IF df IS NULL THEN
		RAISE EXCEPTION ''Reference % not found in d2p_ref_list'', $1;
	END IF;

	IF uc THEN
		tmp := ''BEGIN
			IF NEW."'' || df || ''" = OLD."'' || df || ''" THEN
				RETURN NEW;
			END IF;

			UPDATE "'' ||st|| ''" SET "'' ||sf|| ''" = NEW."''||df||''" WHERE "'' ||sf|| ''" = OLD."''||df||''";
			RETURN NEW;
		END'';
	ELSE
		SELECT INTO tn typname FROM pg_type t, pg_attribute a, pg_class c
			WHERE t.oid = a.atttypid AND a.attrelid = c.oid AND a.attname = sf AND c.relname = st;
		IF tn IS NULL THEN
			tn := st || ''.'' || sf;
			RAISE EXCEPTION ''Type not found for %'', tn;
		END IF;

		tmp := ''DECLARE
			c CURSOR FOR SELECT "'' || sf || ''" FROM "'' || st || ''" WHERE "'' || sf || ''" IS NOT NULL;
			id '' || tn || '';
		BEGIN
			IF NEW."'' || df || ''" = OLD."'' || df || ''" THEN
				RETURN NEW;
			END IF;

			OPEN c;
			LOOP
				FETCH c INTO id;
				IF id IS NULL THEN
					EXIT;
				END IF;
				IF OLD."'' || df || ''" = id THEN
					CLOSE c;
					RAISE EXCEPTION ''''''''Value % found for '' ||st|| ''.'' ||sf|| '''''''''', OLD."'' ||df|| ''";
				END IF;
			END LOOP;
			CLOSE c;
			RETURN NEW;
		END'';
	END IF;
	RETURN tmp;
END' LANGUAGE 'plpgsql';

--
-- Function that generates the code for the source table update check trigger function for a reference
--
CREATE FUNCTION f_d2p_ref_code_f_us(INT) RETURNS TEXT AS '
DECLARE
	idref	ALIAS FOR $1;
	tmp	TEXT;
	st	TEXT;
	sf	TEXT;
	dt	TEXT;
	df	TEXT;
	tn	TEXT;
BEGIN
	SELECT INTO st, sf, dt, df 
		src_table, src_field, dest_table, dest_field FROM d2p_ref_list WHERE id = idref;

	IF df IS NULL THEN
		RAISE EXCEPTION ''Reference % not found in d2p_ref_list'', $1;
	END IF;

	SELECT INTO tn typname FROM pg_type t, pg_attribute a, pg_class c
		WHERE t.oid = a.atttypid AND a.attrelid = c.oid AND a.attname = sf AND c.relname = st;
	IF tn IS NULL THEN
		tn := st || ''.'' || sf;
		RAISE EXCEPTION ''Type not found for %'', tn;
	END IF;

	tmp := ''DECLARE
		c CURSOR FOR SELECT '' || df || '' FROM '' || dt || '';
		id '' ||tn|| '';
	BEGIN
		IF NEW."'' || sf || ''" IS NULL THEN
			RETURN NEW;
		END IF;

		OPEN c;
		LOOP
			FETCH c INTO id;
			IF id IS NULL THEN
				EXIT;
			END IF;
			IF NEW."'' || sf || ''" = id THEN
				CLOSE c;
				RETURN NEW;
			END IF;
		END LOOP;
		CLOSE c;
		RAISE EXCEPTION ''''''''Value % not found for '' || dt || ''.'' || df || '''''''''', NEW."'' || sf || ''";
	END'';
	RETURN tmp;
END' LANGUAGE 'plpgsql';

--
-- Function that generates the code for the destination table delete check trigger function for a reference
--
CREATE FUNCTION f_d2p_ref_code_f_rm(INT) RETURNS TEXT AS '
DECLARE
	idref	ALIAS FOR $1;
	tmp	TEXT;
	st	TEXT;
	sf	TEXT;
	dt	TEXT;
	df	TEXT;
	tn	TEXT;
	dc	BOOL;
BEGIN
	SELECT INTO st, sf, dt, df, dc
		src_table, src_field, dest_table, dest_field, del_cascade FROM d2p_ref_list WHERE id = idref;

	IF df IS NULL THEN
		RAISE EXCEPTION ''Reference % not found in d2p_ref_list'', $1;
	END IF;

	IF dc THEN
		tmp := ''BEGIN
			DELETE FROM "'' ||st|| ''" WHERE "'' ||sf|| ''" = OLD."'' ||df|| ''";
			RETURN OLD;
		END'';
	ELSE
		SELECT INTO tn typname FROM pg_type t, pg_attribute a, pg_class c
			WHERE t.oid = a.atttypid AND a.attrelid = c.oid AND a.attname = sf AND c.relname = st;
		IF tn IS NULL THEN
			tn := st || ''.'' || sf;
			RAISE EXCEPTION ''Type not found for %'', tn;
		END IF;

		tmp := ''DECLARE
			c CURSOR FOR SELECT "'' || sf || ''" FROM "'' || st || ''" WHERE "'' || sf || ''" IS NOT NULL;
			id '' ||tn||'';
		BEGIN
			OPEN c;
			LOOP
				FETCH c INTO id;
				IF id IS NULL THEN
					EXIT;
				END IF;
				IF OLD.'' || df || '' = id THEN
					CLOSE c;
					RAISE EXCEPTION ''''''''Value % found for '' ||st|| ''.'' ||sf|| '''''''''', OLD."'' || df || ''";
				END IF;
			END LOOP;
			CLOSE c;
			RETURN OLD;
		END'';
	END IF;
	RETURN tmp;
END' LANGUAGE 'plpgsql';

--
-- Trigger that creates a function when insertion is required
--
CREATE FUNCTION f_d2p_ref_addfunc () RETURNS OPAQUE AS '
DECLARE
	c	refcursor;
	st	TEXT;
	dt	TEXT;
	tmp	TEXT;
	tmp2	TEXT;
BEGIN
	IF NEW.otype = ''in'' THEN
		SELECT INTO tmp f_d2p_ref_code_f_in(NEW.idref);
	ELSIF NEW.otype = ''us'' THEN
		SELECT INTO tmp f_d2p_ref_code_f_us(NEW.idref);
	ELSIF NEW.otype = ''ud'' THEN
		SELECT INTO tmp f_d2p_ref_code_f_ud(NEW.idref);
	ELSIF NEW.otype = ''rm'' THEN
		SELECT INTO tmp f_d2p_ref_code_f_rm(NEW.idref);
	END IF;

	tmp2 := ''CREATE FUNCTION "f_d2p_ref_auto_'' || NEW.otype || ''_'' || NEW.idref || ''" () RETURNS OPAQUE AS ''''
'' || tmp || '''''' LANGUAGE ''''plpgsql'''';'';
	EXECUTE tmp2;

	IF NEW.otype = ''in'' OR NEW.otype = ''us'' THEN
		-- Create insertion/update trigger on source table and its children
		SELECT INTO st src_table FROM d2p_ref_list WHERE id = NEW.idref;

		IF st IS NULL THEN
			RAISE EXCEPTION ''Reference % not found in d2p_ref_list'', NEW.idref;
		END IF;

		INSERT INTO d2p_ref_triggers(relname, otype, idref) VALUES(st, NEW.otype, NEW.idref);

		OPEN c FOR SELECT child FROM d2p_inh_cache WHERE ancestor = st;
		LOOP
			FETCH c INTO tmp;
			IF tmp IS NULL THEN
				EXIT;
			END IF;

			INSERT INTO d2p_ref_triggers(relname, otype, idref) VALUES(tmp, NEW.otype, NEW.idref);
		END LOOP;
		CLOSE c;
	ELSE
		-- Create update/removal trigger on destination table and its children
		SELECT INTO dt dest_table FROM d2p_ref_list WHERE id = NEW.idref;
		IF dt IS NULL THEN
			RAISE EXCEPTION ''Reference % not found in d2p_ref_list'', NEW.idref;
		END IF;

		INSERT INTO d2p_ref_triggers(relname, otype, idref) VALUES(dt, NEW.otype, NEW.idref);

		OPEN c FOR SELECT child FROM d2p_inh_cache WHERE ancestor = dt;
		LOOP
			FETCH c INTO tmp;
			IF tmp IS NULL THEN
				EXIT;
			END IF;

			INSERT INTO d2p_ref_triggers(relname, otype, idref) VALUES(tmp, NEW.otype, NEW.idref);
		END LOOP;
		CLOSE c;
	END IF;

	RETURN NULL;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_addfunc AFTER INSERT ON d2p_ref_functions
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_addfunc ();

--
-- Trigger that drops a function when a deletion is required
--
CREATE FUNCTION f_d2p_ref_delfunc () RETURNS OPAQUE AS '
BEGIN
	DELETE FROM d2p_ref_triggers WHERE otype = OLD.otype AND idref = OLD.idref;

	EXECUTE ''DROP FUNCTION f_d2p_ref_auto_'' || OLD.otype || ''_'' || OLD.idref || '' () ;'';

	RETURN OLD;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_delfunc BEFORE DELETE ON d2p_ref_functions
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_delfunc ();


--
-- Trigger that creates a trigger on insertion into d2p_ref_triggers
--
CREATE FUNCTION f_d2p_ref_addtrig () RETURNS OPAQUE AS '
DECLARE
	tmp	TEXT;
	op	TEXT;
	wh	TEXT;
	uc	BOOL;
BEGIN
	SELECT INTO uc upd_cascade FROM d2p_ref_list WHERE id = NEW.idref;

	IF NEW.otype = ''in'' THEN
		op := ''INSERT'';
	ELSIF NEW.otype = ''rm'' THEN
		op := ''DELETE'';
	ELSE
		op := ''UPDATE'';
	END IF;

	IF uc AND NEW.otype = ''ud'' THEN
		wh := ''" AFTER '';
	ELSE
		wh := ''" BEFORE '';
	END IF;

	tmp := ''CREATE TRIGGER "t_d2p_ref_auto_'' || NEW.otype || ''_'' || NEW.idref || wh || op || '' ON "'' || NEW.relname || ''"
		FOR EACH ROW EXECUTE PROCEDURE "f_d2p_ref_auto_'' || NEW.otype || ''_'' || NEW.idref || ''" () ;'';
	EXECUTE tmp;
	RETURN NULL;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_addtrig AFTER INSERT ON d2p_ref_triggers
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_addtrig ();


--
-- Trigger that destroys a trigger on deletion from d2p_ref_triggers
--
CREATE FUNCTION f_d2p_ref_deltrig () RETURNS OPAQUE AS '
DECLARE
        tmp     TEXT;
	ok	OID;
BEGIN
	SELECT INTO ok relfilenode FROM pg_class WHERE relname = OLD.relname;
	IF ok IS NOT NULL THEN
		tmp := ''DROP TRIGGER "t_d2p_ref_auto_'' || OLD.otype || ''_'' || OLD.idref || ''" ON "'' || OLD.relname || ''" ;'';
		EXECUTE tmp;
	END IF;
	RETURN OLD;
END' LANGUAGE 'plpgsql';

CREATE TRIGGER t_d2p_ref_deltrig BEFORE DELETE ON d2p_ref_triggers
	FOR EACH ROW EXECUTE PROCEDURE f_d2p_ref_deltrig ();


--
-- Function that updates the triggers
--
CREATE FUNCTION f_d2p_ref_update() RETURNS BOOL AS '
DECLARE
	c	CURSOR FOR SELECT * FROM d2p_ref_list;
	rt	RECORD;
BEGIN
	OPEN c;
	DELETE FROM d2p_ref_list ;
	LOOP
		FETCH c INTO rt;
		IF NOT FOUND THEN
			EXIT;
		END IF;

		INSERT INTO d2p_ref_list(id, src_table, src_field, dest_table, dest_field, del_cascade, upd_cascade) 
			VALUES (rt.id, rt.src_table, rt.src_field, rt.dest_table, rt.dest_field, rt.del_cascade, rt.upd_cascade);
	END LOOP;
	CLOSE c;
	RETURN true;
END' LANGUAGE 'plpgsql';
