package	D2P::Generator::PostgreSQL;
use	strict;
use	base 'D2P::Generator';
use	D2P::Parser;
use	D2P::Utils;


sub	init
{
	my	$this	= shift;

	$this->SUPER::init();
	$this->{config}->{outfile} = "";
	$this->{config}->{sqlfiles} = [];
}


sub	commandLine
{
	my	$this	= shift;

	return	(
		"-o|--output"	=> {
			desc	=> "Output the PostgreSQL script to the specified file.",
			args	=> ["outfile"],
			uniq	=> 1
		},
		"-s|--sql"	=> {
			desc	=> "Add the contents of the specified file at the end of the script. "
					. "May be repeated for each file to include to the script.",
			argl	=> "sqlfiles",
			args	=> ["sqlfile"],
			uniq	=> 0
		}
	);
}


sub	generate
{
	my	$this	= shift;
	my	$parser	= shift;
	my	$schema = $parser->{schema};

	return unless $this->{config}->{outfile};

	$this->{outfile} = $this->{config}->{outfile};
	$this->{sqlfiles} = [ map { $_->{sqlfile} } @{$this->{config}->{sqlfiles}} ];

	## Prepare output
	my	@customCode = ();
	if	(-f $this->{outfile})
	{
		my	($s, @old);

		open(IN, $this->{outfile}) or die "$this->{outfile} : $!\n";
		@old = <IN>;
		close(IN);

		open (SQLOLD, ">$this->{outfile}\~");
		print SQLOLD @old;
		close SQLOLD;

		$s = 0;
		foreach my $l (@old)
		{
			if	(!$s && ($l eq "-- You may insert additionnal SQL code here (DO NOT REMOVE THIS LINE)\n"))
			{
				$s = 1;
			}
			elsif	($s && ($l eq "-- End of additionnal SQL code (DO NOT REMOVE THIS LINE)\n"))
			{
				last;
			}
			elsif	($s)
			{
				push @customCode, $l;
			}
		}
	}
	$this->{customCode} = [@customCode];
	open(SQLOUT, ">$this->{outfile}");

	## Create D2P::Utils instance
	$this->utils(new D2P::Utils());

	$this->createSQLHeader($schema, $parser->file());
	$this->createSQLModules($schema);
	$this->copySQLFiles();
	$this->createSQLFooter($schema);

	close SQLOUT;
	$this->utils(undef);
}


###
## Create SQL header
#
sub	createSQLHeader
{
	my	$this	= shift;
	my	$sch	= shift;
	my	$infile	= shift;

	# Schema header
	my	$title = $sch->header();
	$title =~ s/\n$//;
	my	@lines = split /\n/, $title;
	foreach my $l (@lines)
		{ print SQLOUT "-- $l\n"; }
	print SQLOUT "\n" if @lines;

	# Files
	print SQLOUT "-- Input file  : $infile\n";
	print SQLOUT "-- Output file : " . $this->{outfile} . "\n";
	print SQLOUT "\n";

	# Modules list
	print SQLOUT "-- Defines the following modules : \n";
	foreach my $mod (@{$sch->modules()->{ordered}})
	{
		my	$title = $mod->name();
		$title =~ s/\s+/ /g;

		print SQLOUT "--  $title\n";
	}
	print SQLOUT "\n";

	# Dependencies
	if	(keys %{$sch->extModules()->{byname}})
	{
		print SQLOUT "-- Requires the following modules :\n--  ";
		print SQLOUT join ("\n--  ", keys %{$sch->extModules()->{byname}});
		print SQLOUT "\n\n";
	}
	print SQLOUT "BEGIN TRANSACTION;\n\n\n";
}


###
## Create SQL tables for each included module
#
sub	createSQLModules
{
	my	$this	= shift;
	my	$sch	= shift;

	foreach my $mod (@{$sch->modules()->{ordered}})
	{
		## Print module title
		my	$title = $mod->name();
		$title =~ s/\n/\n-- /g;
		print SQLOUT "----\n-- $title\n----\n\n";

		## Print module comments if any are defined
		my	$hasComments = 0;
		foreach my $cmtEnt (@{$mod->commentsOrdered()})
		{
			next if defined $cmtEnt->relation();

			my	$txt	= $cmtEnt->text();
			$txt =~ s/\n/\n-- /g;
			$txt = "-- $txt\n\n";
			$hasComments = 1;
			print SQLOUT $txt;
		}
		print SQLOUT "\n" if $hasComments;

		## Create sequences
		if	(keys %{$mod->sequences()})
		{
			my	@seqs = keys %{$mod->sequences()};

			print SQLOUT "-- Sequences ($title)\n";
			foreach my $sn (@seqs)
			{
				my	$e	= $mod->sequences()->{$sn};
				my	$snam	= $this->utils()->strToPgId($sn);
				print SQLOUT "CREATE SEQUENCE $snam INCREMENT " . $e->increment() . "\n";
				print SQLOUT "\tMINVALUE " . $e->minvalue() . " MAXVALUE " . $e->maxvalue() . "\n";
				print SQLOUT "\tSTART " . $e->start() . " CACHE " . $e->cache();
				print SQLOUT " CYCLE" if $e->cycle();
				print SQLOUT ";\n";
			}
			print SQLOUT "\n";
		}

		## Create tables
		print SQLOUT "-- Tables ($title)\n\n";
		foreach my $t (@{$mod->tables()->{ordered}})
		{
			my	$x = 0;
			my	@pkey = ();
			my	@references = ();
			my	$tnam = $this->utils()->strToPgId($t->name());

			print SQLOUT "\n--  Table " . $t->name() . "\n";

			## Print table comment if it exists
			if	(defined($t->comment()))
			{
				my	$cmt	= $t->comment();
				my	$cmtEnt	= $cmt->comment();
				my	$txt	= $cmtEnt->text();
				$txt =~ s/\n/\n--   /g;
				$txt = "--   $txt\n";
				print SQLOUT $txt;
			}

			print SQLOUT "CREATE TABLE $tnam (";

			## Table fields
			foreach my $f (@{$t->fields()})
			{
				my	$fnam = $this->utils()->strToPgId($f->fldName());
				$x ? (print SQLOUT ",") : ($x=1);
				print SQLOUT "\n\t$fnam";
				my	$l = 56 - length $fnam;
				if	($l <= 0)
				{
					print SQLOUT " ";
				}
				else
				{
					my	$nt = $l / 8;
					$nt++ if $nt =~ s/\..*$//;
					print SQLOUT "\t" x $nt;
				}

				if	(ref($f) =~ /^D2P::Entity::Field::Serial(::.+)?$/)
				{
					print SQLOUT "INT";

					# NOT NULL constraint
					print SQLOUT " NOT NULL"
						if	($f->notNull());

					# Default value
					print SQLOUT " DEFAULT NEXTVAL('" . $this->utils()->strToPgId($f->sequence()->name()) . "'::TEXT)";
				}
				else
				{
					print SQLOUT $f->fldType();

					# NOT NULL constraint
					print SQLOUT " NOT NULL"
						if	($f->notNull());

					# Default value
					print SQLOUT " DEFAULT " . $f->default()
						if	$f->default() ne "";

					# Prepare reference if required
					if	($f->reference())
					{
						my	$field	= $this->utils()->strToPgId($f->reference()->fldName());
						my	$table	= $this->utils()->strToPgId($f->reference()->table()->name());
						if	(!$f->relation()->usePostgres())
						{
							my	$tmpStr = "'$tnam', '$fnam', '$table', '$field', ";

							$tmpStr .= $f->relation()->deleteCascade() ? "true" : "false";
							$tmpStr .= ", ";
							$tmpStr .= $f->relation()->updateCascade() ? "true" : "false";
							push @references, $tmpStr;
						}
						else
						{
							print SQLOUT " REFERENCES $table\($field\) ON UPDATE ";
							print SQLOUT ($f->relation()->updateCascade() ? "CASCADE" : "NO ACTION");
							print SQLOUT " ON DELETE ";
							print SQLOUT ($f->relation()->deleteCascade() ? "CASCADE" : "NO ACTION");
						}
					}
				}

				## Print field comment if it is defined
				if	(defined($f->comment()))
				{
					my	$cmt	= $f->comment();
					my	$cmtEnt	= $cmt->comment();
					my	$txt	= $cmtEnt->text();
					$txt =~ s/\n/\n\t\t-- /g;
					$txt = "\n\t\t-- $txt";
					print SQLOUT $txt;
				}
			}

			# Table primary key
			if	(@{$t->pkey()})
			{
				my	@opk = map { $this->utils()->strToPgId($_->fldName()) } (@{$t->pkey()});
				print SQLOUT "," if @{$t->fields()};
				print SQLOUT "\n\tPRIMARY KEY(" . join(",", @opk) . ")";
			}
			print SQLOUT "\n)";

			# Table ancestors
			my	@d2pInheritance = ();
			if	(@{$t->ancestors()})
			{
				my	$y = 0;

				print SQLOUT " INHERITS (";
				foreach my $inhRel (@{$t->ancestors()})
				{
					my	$super = $inhRel->to();
					$y ? (print SQLOUT ", ") : ($y = 1);
					print SQLOUT (my $snam = $this->utils()->strToPgId($super->name()));
					# D2P inheritance ignore list
					push @d2pInheritance, "INSERT INTO d2p_inh_ignore(ancestor,child)"
						. " VALUES ('$snam', '$tnam');"
							if $inhRel->usePostgres();
				}
				print SQLOUT ")";
			}
			print SQLOUT ";\n\n";

			# Create indices for this table
			foreach my $i ($t->getIndices())
			{
				## Skip empty indices
				my	@fld = $i->getFields();
				next unless @fld;

				## Print index comment if it is defined
				if	(defined($i->comment()))
				{
					my	$cmt	= $i->comment();
					my	$cmtEnt	= $cmt->comment();
					my	$txt	= $cmtEnt->text();
					$txt =~ s/\n/\n-- /g;
					$txt = "-- $txt\n";
					print SQLOUT $txt;
				}

				## Print the index definition
				print SQLOUT "CREATE " . ($i->isUnique() ? 'UNIQUE ' : '') . "INDEX ";
				print SQLOUT $this->utils()->strToPgId($i->{name}) . " ON $tnam";
				#$i->{type} && print SQLOUT " USING " . $i->{type}; FIXME no longer supported
				print SQLOUT " ("
					. join(',', map { $this->utils()->strToPgId($_->fldName()); } @fld)
					. ");\n";
			}
			print SQLOUT "\n" if $t->getIndices();

			# Table contents if available
			foreach my $cnt (@{$t->{contents}})
			{
				print SQLOUT "INSERT INTO $tnam("
					. join(",", map { $this->utils()->strToPgId($_) } keys(%$cnt)) . ")\n";
				print SQLOUT "\tVALUES (" . join(",", values(%$cnt)) . ");\n";
			}
			print SQLOUT "\n" if @{$t->{contents}};

			# D2P SQL library stuff for list and references
			$tnam =~ s/\"//g;
			## Table list
			print SQLOUT "INSERT INTO d2p_lst_list(d2p_name, dia_name) VALUES ('$tnam', '$t->{name}');\n";
			## References
			foreach my $ref (@references)
			{
				$ref =~ s/\"//g;
				print SQLOUT "INSERT INTO d2p_ref_list(src_table, src_field, "
					. "dest_table, dest_field, del_cascade, upd_cascade)\n";
				print SQLOUT "\tVALUES ($ref);\n";
			}
			## Postgres inheritance
			print SQLOUT join("\n", @d2pInheritance) . "\n" if @d2pInheritance;
			print SQLOUT "\n";
		}
		print SQLOUT "\n";
	}
}


###
## Copy additional SQL files
#
sub	copySQLFiles
{
	my	$this	= shift;
	return unless @{$this->{sqlfiles}};

	print SQLOUT "\n\n-- \n-- ADDITIONAL SQL FILES\n-- \n";
	foreach my $f (@{$this->{sqlfiles}})
	{
		unless	(open(SFIC, $f))
		{
			print STDERR "WARNING : could not open SQL file '$f'\n";
			next;
		}
		print SQLOUT "\n-- \n-- SQL code from $f\n-- \n\n";
		my @cff;
		@cff = <SFIC>;
		close SFIC;
		print SQLOUT @cff;
	}
}


###
## Create SQL footer
#
sub	createSQLFooter
{
	my	$this	= shift;

	print SQLOUT "\n\n-- You may insert additionnal SQL code here (DO NOT REMOVE THIS LINE)\n";
	if	(@{$this->{customCode}})
	{
		print SQLOUT @{$this->{customCode}};
	}
	else
	{
		print SQLOUT "\n" x 4;
	}
	print SQLOUT "-- End of additionnal SQL code (DO NOT REMOVE THIS LINE)\n\n";
	print SQLOUT "SELECT f_d2p_inh_update();\nSELECT f_d2p_ref_update();\n\n";
	print SQLOUT "COMMIT TRANSACTION;\n";
	print SQLOUT "-- SQL code generated by dia2postgres.pl\n";
	print SQLOUT "-- $Main::VERSION\n"; 
	print SQLOUT "-- on " . `date`;
}




1;
