package D2P::Parser;
use base 'D2P::Base';

# D2P Entity classes
use	D2P::Entity::Schema;
use	D2P::Entity::Module;
use	D2P::Entity::Comment;
use	D2P::Entity::Table;
use	D2P::Entity::Field;
use	D2P::Entity::Field::Enum;
use	D2P::Entity::Field::Serial;
use	D2P::Entity::Index;
use	D2P::Entity::Relation::Comment;
use	D2P::Entity::Relation::Inheritance;
use	D2P::Entity::Relation::Reference;

# XML Parser and compression library
use	XML::Parser;
eval "use Compress::Zlib;";
my	$zlibAvailable = ($@) ? 0 : 1;


###
## Read the XML file and parse it
###
sub	init
{
	my	$this	= shift;
	my	$file	= shift;

	die "ERROR : the '-i' option is now mandatory\n"
		unless $file;

	$this->{file} = $file;
	my	$parser	= new XML::Parser(Style => 'Tree');
	my	$xml	= "";

	if	($zlibAvailable)
	{
		my	($zStream, $zStatus)	= deflateInit();
		my	$gz			= gzopen($file, "rb");
		if	($gz)
		{
			my	$buffer;
			while	($gz->gzread($buffer) > 0)
			{
				$xml .= $buffer;
			}
			chomp $xml;
		}
	}

	if	($xml eq "")
	{
		open(INFILE, $file)
			or die "ERROR : could not open '$file'\n";
		my @all = <INFILE>;
		close INFILE;
		$xml = join("", @all);
	}

	die "ERROR : '$file' is not an XML file.\n"
		unless $xml =~ /^\s*<\?\s*xml/;

	eval '$this->{xmldata} = $parser->parse($xml);';
	die "ERROR : XML data from '$file' could not be parsed\n"
		if $@;
}


###
## Transform the XML data read from the input file into entities
###
sub	createStructure
{
	my	$this	= shift;
	my	@data	= @{$this->{xmldata}};
	my	($schema, $tagPrefix);

	return if defined($this->{schema});
	$this->{schema} = $schema = new D2P::Entity::Schema();

	my	$tmp = $data[0];
	$tagPrefix = ($tmp =~ /^dia:/) ? "dia:" : "";
	die "ERROR : 'diagram' tag not found\n"
		if	($tmp ne $tagPrefix . "diagram");

	my	$ptr = 1;
	@data = @{$data[1]};
	while	($ptr < @data)
	{
		$ptr += 2 while ($ptr < @data && $data[$ptr] ne $tagPrefix . "layer");
		if	($ptr < @data)
		{
			my	@layerData = @{$data[$ptr + 1]};
			$this->analyzeLayer($tagPrefix, @layerData[1 .. (@layerData+0)])
				if	(@layerData > 2);
			$ptr += 2;
		}
	}

	$this->{schema}->postprocess();
}


###
## Analyse a layer from the diagram
###
sub	analyzeLayer
{
	my	$this	= shift;
	my	$tp	= shift;
	my	@lData	= @_;

	for	(my $i=0;$i<@lData - 1;$i+=2)
	{
		if	($lData[$i] eq $tp."group")
		{
			my	@groupData = @{$lData[$i + 1]};
			$this->analyzeGroup($tp, @groupData[1 .. (@groupData + 0)])
				if	(@groupData > 2);
		}
		elsif	($lData[$i] eq $tp."object")
		{
			my	$obj = $this->readDIAObject($tp, @{$lData[$i + 1]});
			my	$ent = $this->analyzeObject($obj);
			if	(!ref($ent))
			{
				print STDERR "WARNING : unknown object type '" . $obj->{type} . "' at layer level\n";
			}
			elsif	(ref($ent) !~ /^D2P::Entity::Relation::.+$/)
			{
				print STDERR "WARNING : unexpected object type '" . $obj->{type} . "' at layer level\n";
			}
			else
			{
				$this->{schema}->addRelation($ent);
			}
		}
		elsif	($lData[$i] ne "0")
		{
			print STDERR "WARNING : tag '" . $lData[$i] . "' ignored at layer level\n";
		}
	}
}


###
## Analyze and create data from a DIA group
###
sub	analyzeGroup
{
	my	$this	= shift;
	my	$tp	= shift;	# Tag prefix
	my	@oData	= @_;		# Original group data

	# Get rid of nested groups and text data
	my	(@data, $i);
	my	$str	= $tp . "group";
	while	(grep(/^$str$/, @oData) || grep(/^0$/, @oData))
	{
		@data = ();
		for	($i=0;$i<@oData;$i+=2)
		{
			if	($oData[$i] eq $tp."object")
			{
				@data = (@data, @oData[$i .. ($i + 1)]);
			}
			elsif	($oData[$i] eq $str)
			{
				my	@gd = @{$oData[$i + 1]};
				@data = (@data, @gd[1 .. (@gd - 1)]) unless (@gd < 3);
			}
		}
		@oData = @data;
	}

	# Transform objects into hash tables
	my	@objs = ();
	for	($i=1;$i<@oData;$i+=2)
	{
		push @objs, $this->readDIAObject($tp, @{$oData[$i]});
	}

	die "ERROR : a group should contain at least two objects\n"
		if	(@objs < 2);

	# Find the group container (i.e. box or large package) and try to find text
	my	$gc	= undef;
	my	$txt	= undef;
	my	@nobjs	= ();
	for $i (0 .. (@objs - 1))
	{
		my $dgc = defined $gc;
		my $dtx = defined $txt;
		if	(	($objs[$i]->{type} eq "Standard - Box")
			||	($objs[$i]->{type} eq "UML - LargePackage")
			)
		{
			die "ERROR : more than one container found in a group\n"
				if	(	$dgc
					&&	(	($gc->{type} eq "UML - LargePackage")
						||	($objs[$i]->{type} eq "UML - LargePackage")
						)
					);
			$gc = $objs[$i] unless $dgc;
		}
		elsif	($objs[$i]->{type} eq "Standard - Text")
		{
			die "ERROR : more than one text found in a group\n" if $dtx;
			$txt = $objs[$i];
		}
		else
		{
			push @nobjs, $objs[$i];
		}
	}
	die "ERROR : group container not found\n"
		unless defined $gc;
	@objs = @nobjs;

	# Validate group data ; a group should include at least :
	#  * one standard box and a standard text
	#  * one UML large package and a class
	my	$module	= undef;
	if	($gc->{type} =~ /^Sta/ && !@objs && defined($txt))
	{
		die "ERROR : schema header found twice\n"
			if ($this->{schema}->header() ne "");
		$this->{schema}->header(
			$txt->{attributes}->{text}->{value}->[0]->{attributes}->{string}->{value}
		);
	}
	elsif	($gc->{type} =~ /^UML/)
	{
		die "ERROR : text found inside an external module (DIA UML Large Package)\n"
			if	defined($txt);
		die "ERROR : empty external module\n"
			if	!@objs;

		$module = new D2P::Entity::Module($gc->{attributes}->{name}->{value});
		$module->isExternal(1);
		$module->diaId($gc->{id});
		$this->analyzeExternalModule($module, @objs);
		$module->preprocess();
	}
	else
	{
		die "ERROR : invalid module found\n"
			unless	(@objs && defined($txt));

		my	$n = $txt->{attributes}->{text}->{value}->[0]->{attributes}->{string}->{value};
		$module = new D2P::Entity::Module($n);
		$module->diaId($gc->{id});
		$this->analyzeModule($module, @objs);
		$module->preprocess();
	}

	$this->{schema}->addModule($module) if defined($module);
}


###
## Analyze objects for an external module
###
sub	analyzeExternalModule
{
	my	($this, $mod)	= (shift, shift);
	my	@objects	= @_;

	# Convert DIA objects into entities
	foreach	my $o (@objects)
	{
		my	$ent = $this->analyzeObject($o);
		if	(defined($ent) && ref($ent) && ref($ent) =~ /^D2P::Entity::/)
		{
			my	$etype = ref($ent);
			$etype =~ s/^D2P::Entity:://;

			if	($etype =~ /^Table(::.*)?$/)
			{
				$mod->addTable($ent);
			}
			else
			{
				print STDERR "WARNING : Object type '" . $o->{type} . "' not allowed inside external module\n";
			}
		}
	}
}


###
## Analyze objects for a module
###
sub	analyzeModule
{
	my	($this, $mod)	= (shift, shift);
	my	@objects	= @_;

	# Convert DIA objects into entities
	foreach	my $o (@objects)
	{
		my	$ent = $this->analyzeObject($o);
		if	(defined($ent) && ref($ent) && ref($ent) =~ /^D2P::Entity::/)
		{
			my	$etype = ref($ent);
			$etype =~ s/^D2P::Entity:://;

			if	($etype =~ /^Table(::.*)?$/)
			{
				$mod->addTable($ent);
			}
			elsif	($etype =~ /^Relation::.+$/)
			{
				$mod->addRelation($ent);
			}
			elsif	($etype =~ /^Comment(::.*)?$/)
			{
				$mod->addComment($ent);
			}
			else
			{
				print STDERR "WARNING : Object type '" . $o->{type} . "' not allowed inside internal module\n";
			}
		}
	}
}


###
## Analyze an object and create entity data from it
###
sub	analyzeObject
{
	my	$this	= shift;
	my	$obj	= shift;
	my	$ent	= undef;

	if	($obj->{type} eq "UML - Class")
	{
		$ent = $this->analyzeTable($obj);
	}
	elsif	(	$obj->{type} eq "Standard - Line"
		||	$obj->{type} eq "Standard - Arc"
		||	$obj->{type} eq "Standard - ZigZagLine"
		)
	{
		$ent = $this->analyzeStandardLine($obj);
	}
	elsif	($obj->{type} eq "UML - Note")
	{
		$ent = $this->analyzeUMLNote($obj);
	}
	elsif	($obj->{type} eq "UML - Implements")
	{
		$ent = $this->analyzeUMLComment($obj);
	}
	elsif	($obj->{type} eq "UML - Dependency")
	{
		$ent = $this->analyzeUMLReference($obj);
	}
	elsif	(	$obj->{type} eq "UML - Realizes"
		||	$obj->{type} eq "UML - Generalization"
		)
	{
		$ent = $this->analyzeUMLInheritance($obj);
	}
	else
	{
		print STDERR "WARNING : unsupported object type '" . $obj->{type} . "'\n";
	}

	return	$ent;
}


###
## Convert an UML class object into a Table entity
###
sub	analyzeTable
{
	my	$this	= shift;
	my	$obj	= shift;

	# Create Table entity
	my	$t	= new D2P::Entity::Table($obj->{attributes}->{name}->{value});
	$t->diaId($obj->{id});

	# Read fields
	if	($obj->{attributes}->{attributes}->{type} eq 'composite')
	{
		my	@fList = @{$obj->{attributes}->{attributes}->{value}};
		foreach my $c (@fList)
		{
			next unless $c->{type} eq 'umlattribute';
			my	%attr = %{$c->{attributes}};
			my	$cl = "D2P::Entity::Field";
			if	($attr{type}->{value} =~ /^\s*\{[^\}]*\}\s*$/)
			{
				$cl .= "::Enum";
			}
			elsif	($attr{type}->{value} =~ /^\s*serial\s*/i)
			{
				$cl .= "::Serial";
			}
			my	$field = $cl->new($attr{name}->{value}, $attr{type}->{value});
			$field->default($attr{value}->{value}) if ($attr{value}->{value} ne "");
			$field->notNull(1) if ($attr{class_scope}->{value} eq 'true' || $attr{visibility}->{value} eq "0");

			#print "Field " . $t->name() . "." . $field->fldName() . " " . $field->fldType();
			#$field->default() ne "" && print " DEFAULT " . $field->default();
			#$field->notNull() && print " NOT NULL";
			#print "\n";

			$t->addField($field, ($attr{visibility}->{value} eq "0"));
		}
	}

	# Read indices
	if	($obj->{attributes}->{operations}->{type} eq 'composite')
	{
		my	@iList = @{$obj->{attributes}->{operations}->{value}};
		foreach my $c (@iList)
		{
			next unless $c->{type} eq 'umloperation';

			my	%attr	= %{$c->{attributes}};
			unless	($attr{parameters}->{type} eq 'composite')
			{
				print STDERR "WARNING : empty operation ignored for table " . $t->name() . "\n";
				next;
			}

			my	$cl	= "D2P::Entity::Index";
			my	$index	= $cl->new();
			$index->isUnique($attr{visibility}->{value} ne '0');

			my	@fList	= @{$attr{parameters}->{value}};
			my	@iFList	= ();
			foreach my $c2 (@fList)
			{
				next unless $c2->{type} eq 'umlparameter';
				my	$fn = $c2->{attributes}->{name}->{value};
				my	$f = $t->fieldByName($fn);
				if	(defined($f))
				{
					push @iFList, $fn;
					$index->addField($f);
				}
				else
				{
					push @iFList, "I$fn";
					$index->addUnresolvedField($fn);
				}
			}

			my	$name	= $attr{name}->{value};
			$name = "i_" . $t->name() . "_" . join("_", @iFList)
				if	($name eq "");
			$index->name($name);

			#print "Index " . $t->name() . "." . $index->name();
			#$index->isUnique() && print " UNIQUE";
			#print " (" . join(", ", @iFList) . ")";
			#print "\n";

			$t->addIndex($index);
		}
	}

	return	$t;
}


###
## Convert a standard line DIA object into a Relation entity
###
sub	analyzeStandardLine
{
	my	$this	= shift;
	my	$obj	= shift;
	my	$rel	= undef;
	my	$eAr	= $obj->{attributes}->{end_arrow}->{value};

	if	($eAr eq "2")
	{
		$rel = new D2P::Entity::Relation::Inheritance(
			$obj->{connections}->[0]->{destId},
			$obj->{connections}->[1]->{destId}
		);
	}
	elsif	(grep(/^$eAr$/, ("1", "6", "7", "12")))
	{
		$rel = new D2P::Entity::Relation::Reference(
			$obj->{connections}->[0]->{destId},
			$obj->{connections}->[1]->{destId},
			$obj->{connections}->[0]->{destCnx},
			$obj->{connections}->[1]->{destCnx}
		);
		if	($eAr eq "6")
		{
			$rel->deleteCascade(1);
		}
		elsif	($eAr eq "7")
		{
			$rel->updateCascade(1);
		}
		elsif	($eAr eq "12")
		{
			$rel->deleteCascade(1);
			$rel->updateCascade(1);
		}
	}
	else
	{
		print STDERR "WARNING : ignored relation with unknown type\n";
	}
	if	(defined($rel))
	{
		# Set entity's DIA identifier
		$rel->diaId($obj->{id});

		# Determine whether the user wants D2P or Postgres relations
		$rel->usePostgres(1)
			if ($obj->{attributes}->{line_style}->{value} eq '4');
	}
	return	$rel;
}


###
## Convert an UML realisation or generalisation DIA object into a
## Relation::Inheritance entity
###
sub	analyzeUMLInheritance
{
	my	$this	= shift;
	my	$obj	= shift;
	my	$rel	= undef;

	$rel = new D2P::Entity::Relation::Inheritance(
		$obj->{connections}->[1]->{destId},
		$obj->{connections}->[0]->{destId}
	);

	# Set entity's DIA identifier
	$rel->diaId($obj->{id});

	# Determine whether the user wants D2P or Postgres relations
	$rel->usePostgres(1)
		if ($obj->{attributes}->{stereotype}->{value} eq 'pg');
	return	$rel;
}


###
## Convert an UML dependency DIA object into a Relation::Reference entity
###
sub	analyzeUMLReference
{
	my	$this	= shift;
	my	$obj	= shift;
	my	$rel	= undef;

	$rel = new D2P::Entity::Relation::Reference(
		$obj->{connections}->[0]->{destId},
		$obj->{connections}->[1]->{destId},
		$obj->{connections}->[0]->{destCnx},
		$obj->{connections}->[1]->{destCnx}
	);

	# Set entity's DIA identifier
	$rel->diaId($obj->{id});

	# Determine whether the user wants D2P or Postgres relations
	$rel->usePostgres(1)
		if ($obj->{attributes}->{stereotype}->{value} eq 'pg');
	return	$rel;
}


###
## Convert an UML note DIA object into a Comment entity
###
sub	analyzeUMLNote
{
	my	$this	= shift;
	my	$obj	= shift;
	my	$ent	= undef;

	$ent = new D2P::Entity::Comment();
	# Set entity's DIA identifier
	$ent->diaId($obj->{id});
	# Set comment text
	$ent->text($obj->{attributes}->{text}->{value}->[0]->{attributes}->{string}->{value});

	return	$ent;
}


###
## Convert an UML implementation DIA object into a Comment relation entity
###
sub	analyzeUMLComment
{
	my	$this	= shift;
	my	$obj	= shift;
	my	$ent	= undef;

	$ent = new D2P::Entity::Relation::Comment(
		$obj->{connections}->[0]->{destId},
		$obj->{connections}->[1]->{destId},
		$obj->{connections}->[0]->{destCnx}
	);

	# Set entity's DIA identifier
	$ent->diaId($obj->{id});

	return	$ent;
}


###
## Read a DIA object into a hash table
###
sub	readDIAObject
{
	my	$this	= shift;
	my	$tp	= shift;	# Tag prefix
	my	@oData	= @_;
	my	$hash	= {};

	# Read type and DIA identifier
	$hash->{type} = $oData[0]->{type};
	$hash->{id} = $oData[0]->{id};

	# Read object attributes
	$hash->{attributes} = {};
	$this->readDIAAttributes($tp, $hash->{attributes}, @oData[1 .. (@oData - 1)])
		if	(@oData > 2);

	# Read object connections
	$hash->{connections} = [];
	my	$i = 1;
	while	($i < @oData)
	{
		if	($oData[$i] eq $tp."connections")
		{
			my	@cnLst = @{$oData[$i + 1]};
			@cnLst = @cnLst[1 .. (@cnLst - 1)];

			for	(my $j=0;$j<@cnLst;$j+=2)
			{
				next unless $cnLst[$j] eq $tp."connection";
				my	$cid	= int($cnLst[$j + 1]->[0]->{handle});
				next if defined $hash->{connections}->[$cid];
				my	$cdst	= {
					destId	=> $cnLst[$j + 1]->[0]->{to},
					destCnx	=> $cnLst[$j + 1]->[0]->{connection}
				};
				$hash->{connections}->[$cid] = $cdst;
			}

			last;
		}
		$i += 2;
	}

	return	$hash;
}


###
## Reads object or composite attributes from the XML data tree
###
sub	readDIAAttributes
{
	my	$this	= shift;
	my	$tp	= shift;	# Tag prefix
	my	$hr	= shift;	# Reference to the attributes hash table
	my	@list	= @_;		# Tag list from the diagram

	for	(my $i=1;$i<@list;$i+=2)
	{
		next if ($list[$i - 1] ne $tp."attribute");

		# Get attribute name
		my	@attr	= @{$list[$i]};
		my	$an	= $attr[0]->{name};

		# Find attribute contents
		my	$j	= 1;
		$j += 2 while ($attr[$j] eq '0');

		# Get attribute type
		my	$at	= $attr[$j];
		my	$av	= undef;
		$at =~ s/^$tp// if ($tp ne '');

		# Special handling is required when a composite is found
		if	($at eq 'composite')
		{
			$av = [];
			while	($j < @attr)
			{
				if	($attr[$j] eq $tp.'composite' && @{$attr[$j + 1]} > 2)
				{
					my	@lsub = @{$attr[$j + 1]};
					my	$elem = {
						type		=> $lsub[0]->{type},
						attributes	=> {}
					};
					$this->readDIAAttributes($tp, $elem->{attributes}, @lsub[1 .. (@lsub - 1)]);
					push @{$av}, $elem;
				}
				$j += 2;
			}
		}
		# Strings have their value passed as text instead of tag attribute
		elsif	($at eq 'string')
		{
			if	(@{$attr[$j + 1]} == 1)
			{
				$av = "";
			}
			elsif	($attr[$j + 1]->[1] eq '0')
			{
				$av = $attr[$j + 1]->[2];
				$av =~ s/^\#//;
				$av =~ s/\#$//;
			}
			else
			{
				die "ERROR : invalid DIA string found\n";
			}
		}
		# Other type ; value may be found in the 'val' tag attribute
		else
		{
			$av = $attr[$j + 1]->[0]->{val};
		}
		# Store attribute
		$hr->{$an} = {
			type	=> $at,
			value	=> $av
		};
	}
}


1;
